# -*- encoding: utf-8 -*-
from py3o.template.data_struct import Py3oDataError
from py3o.template.helpers import Py3oConvertor

import unittest
import os

import lxml.etree
import pkg_resources
import six

from pyjon.utils import get_secure_filename

from py3o.template.main import move_siblings, detect_keep_boundary, Template

from py3o.template.data_struct import (
    Py3oModule,
    Py3oName,
    Py3oArray,
)

if six.PY3:
    # noinspection PyUnresolvedReferences
    from unittest.mock import Mock
elif six.PY2:
    # noinspection PyUnresolvedReferences
    from mock import Mock


class TestHelpers(unittest.TestCase):

    def tearDown(self):
        pass

    def setUp(self):
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_example_template.odt'
        )

        outname = get_secure_filename()

        self.reference_template = Template(template_name, outname)
        os.unlink(outname)

    def test_move_1(self):
        """test that siblings are properly moved without keeping boundaries"""
        template_one_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/move_one.xml'
        )
        test_template_one = lxml.etree.parse(template_one_name)
        start = test_template_one.find('mystruct/start')
        end = test_template_one.find('mystruct/end')
        new_ = lxml.etree.Element("finishcontainer")

        move_siblings(start, end, new_)

        result_s = lxml.etree.tostring(
            test_template_one,
            pretty_print=True,
        ).decode('utf-8')

        expected_result = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/move_one_result.xml'
            )
        ).read()

        expected_result = expected_result.replace(
            '\n', '').replace(' ', '').strip()

        result_s = result_s.replace('\n', '').replace(' ', '').strip()

        assert result_s == expected_result

    def test_move_2(self):
        """start.tail is correctly copied without keeping boundaries"""
        template_two_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/move_two.xml'
        )
        test_template_two = lxml.etree.parse(template_two_name)
        start = test_template_two.find('mystruct/start')
        end = test_template_two.find('mystruct/end')
        new_ = lxml.etree.Element('finishcontainer')

        move_siblings(start, end, new_)

        result_s = lxml.etree.tostring(
            test_template_two,
            pretty_print=True,
        ).decode('utf-8')

        expected_result = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/move_two_result.xml'
            )
        ).read()

        expected_result = expected_result.replace(
            '\n', '').replace(' ', '').strip()

        result_s = result_s.replace('\n', '').replace(' ', '').strip()

        assert result_s == expected_result

    def test_move_keep_boundaries(self):
        """test that siblings are properly moved keeping boundaries"""
        template_three_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/move_three.xml'
        )
        test_template_three = lxml.etree.parse(template_three_name)
        start = test_template_three.find('mystruct/start')
        end = test_template_three.find('mystruct/end')

        new_ = lxml.etree.Element('finishcontainer')

        move_siblings(
            start, end, new_,
            keep_start_boundary=True,
            keep_end_boundary=True,
        )

        result_s = lxml.etree.tostring(
            test_template_three,
            pretty_print=True,
        ).decode('utf-8')

        expected_result = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/move_three_result.xml'
            )
        ).read()

        expected_result = expected_result.replace(
            '\n', '').replace(' ', '').strip()

        result_s = result_s.replace('\n', '').replace(' ', '').strip()

        assert result_s == expected_result

    def test_move_2_keep_boundaries(self):
        """test that start.tail is correctly copied keeping boundaries"""
        template_four_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/move_four.xml'
        )
        test_template_four = lxml.etree.parse(template_four_name)
        start = test_template_four.find('mystruct/start')
        end = test_template_four.find('mystruct/end')

        new_ = lxml.etree.Element('finishcontainer')

        move_siblings(
            start, end, new_,
            keep_start_boundary=True,
            keep_end_boundary=True,
        )

        result_s = lxml.etree.tostring(
            test_template_four,
            pretty_print=True,
        ).decode('utf-8')

        expected_result = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/move_four_result.xml'
            )
        ).read()

        expected_result = expected_result.replace(
            '\n', '').replace(' ', '').strip()

        result_s = result_s.replace('\n', '').replace(' ', '').strip()

        assert result_s == expected_result

    def test_get_user_variables(self):
        source_odt_filename = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_example_template.odt'
        )
        outfilename = get_secure_filename()

        template = Template(source_odt_filename, outfilename)

        user_vars = template.get_user_variables()

        expected_vars = [
            'line.val1',
            'line.val2',
            'line.val3',
            'item.Amount',
            'item.Currency',
            'item.InvoiceRef',
            'document.total',
        ]
        assert set(user_vars) == set(expected_vars)

    def test_convertor(self):
        source_odt_filename = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_if_parser.odt'
        )
        outfilename = get_secure_filename()

        template = Template(source_odt_filename, outfilename)

        user_vars = template.get_user_variables()
        expressions = template.get_all_user_python_expression()
        py_expression = template.convert_py3o_to_python_ast(expressions)
        convertor = Py3oConvertor()
        data_struct = convertor(py_expression)

        assert 'objects' in data_struct
        objs = data_struct['objects']
        assert 'company_label' in objs
        assert 'name' in objs
        assert isinstance(objs['company_label'], Py3oName)
        assert isinstance(objs['name'], Py3oName)
#         assert data_struct == Py3oModule(
#             {
#                 'objects': Py3oArray(
#                             {
#                                 'company_label': Py3oName({}),
#                                 'name': Py3oName({})
#                             }
#                 )
#             }
#         )
        expected_vars = [
            'registration.name',
            'registration.company_label',
        ]
        assert set(user_vars) == set(expected_vars)

    def test_get_user_instructions(self):
        source_odt_filename = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_example_template.odt'
        )
        outfilename = get_secure_filename()

        template = Template(source_odt_filename, outfilename)

        user_instructions = template.get_user_instructions()

        expected_vars = [
            'for="line in items"',
            '/for',
            'for="item in items"',
            'if="item.InvoiceRef==\'#1234\'"',
            '/if',
            '/for',
        ]
        assert set(user_instructions) == set(expected_vars)

    def test_calc(self):
        """Test date source extraction in ods files"""
        source_ods_filename = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_simple_calc.ods'
        )
        outfilename = get_secure_filename()

        template = Template(source_ods_filename, outfilename)

        expressions = template.get_all_user_python_expression()
        expected_expressions = [
            'for="item in items"',
            'item.col1',
            'item.col2',
            'item.col3',
            'item.col4',
            '/for',
        ]
        self.assertEqual(expressions, expected_expressions)

        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        user_data = {'items': [
            Mock(col0='0', col1=1, col2=2.0, col3='?', col4='!'),
            Mock(col0=0, col1=1.0, col2='2', col3='?', col4='!'),
            Mock(col0=0.0, col1='1', col2=2, col3='?', col4='!'),
        ]}

        json_dict = res.render(user_data)
        self.assertEqual(json_dict, {'items': [
            dict(col1=1, col2=2.0, col3='?', col4='!'),
            dict(col1=1.0, col2='2', col3='?', col4='!'),
            dict(col1='1', col2=2, col3='?', col4='!'),
        ]})

    def test_detect_boundary_false(self):
        """boundary detection should say no!!"""

        source_xml_filename = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/keepboundary_detection_false.xml'
        )

        test_xml = lxml.etree.parse(source_xml_filename)
        starts, ends = self.reference_template.find_instructions(
            [test_xml], self.reference_template.namespaces
        )
        for start, base in starts:
            end = ends[id(start)]
            keep_start, keep_end = detect_keep_boundary(
                start, end,  self.reference_template.namespaces
            )
            assert keep_start is False
            assert keep_end is False

    def test_detect_boundary_true(self):
        """boundary detection should say yes!!"""

        source_xml_filename = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/keepboundary_detection_true.xml'
        )

        test_xml = lxml.etree.parse(source_xml_filename)
        starts, ends = self.reference_template.find_instructions(
            [test_xml], self.reference_template.namespaces
        )
        for index, (start, base) in enumerate(starts):
            end = ends[id(start)]
            keep_start, keep_end = detect_keep_boundary(
                start, end,  self.reference_template.namespaces
            )
            if index == 0:
                assert keep_start is False
                assert keep_end is True

            else:
                assert False, "We should find one single link"

    def test_move_siblings_1(self):
        template_xml = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/move_siblings.xml'
        )
        test_xml = lxml.etree.parse(template_xml)
        starts, ends = self.reference_template.find_instructions(
            [test_xml], self.reference_template.namespaces
        )

        assert len(starts) == 1
        assert len(ends) == 1

        start, _ = starts[0]
        end = ends[id(start)]

        keep_start, keep_end = detect_keep_boundary(
            start, end,  self.reference_template.namespaces
        )

        assert keep_start is True
        assert keep_end is False

        new_ = lxml.etree.Element('finishcontainer')

        start_parent = start.getparent()
        end_parent = end.getparent()

        start.getparent().remove(start)
        end.getparent().remove(end)

        move_siblings(
            start_parent, end_parent, new_,
            keep_start_boundary=keep_start,
            keep_end_boundary=keep_end,
        )

        result_a = lxml.etree.tostring(
            test_xml,
            pretty_print=True,
        ).decode('utf-8')

        result_e = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/move_siblings_result_1.xml'
            )
        ).read()

        result_a = result_a.replace("\n", "").replace(" ", "")
        result_e = result_e.replace("\n", "").replace(" ", "")

        assert result_a == result_e

    def test_move_siblings_2(self):
        template_xml = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/move_siblings.xml'
        )
        test_xml = lxml.etree.parse(template_xml)
        starts, ends = self.reference_template.find_instructions(
            [test_xml], self.reference_template.namespaces
        )

        assert len(starts) == 1
        assert len(ends) == 1

        start, base = starts[0]
        end = ends[id(start)]

        keep_start, keep_end = detect_keep_boundary(
            start, end,  self.reference_template.namespaces
        )

        assert keep_start is True
        assert keep_end is False

        self.reference_template.handle_link(start, base, end)
        result_a = lxml.etree.tostring(
            test_xml,
            pretty_print=True,
        ).decode('utf-8')

        result_e = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/move_siblings_result_2.xml'
            )
        ).read()

        result_a = result_a.replace("\n", "").replace(" ", "")
        result_e = result_e.replace("\n", "").replace(" ", "")

        assert result_a == result_e

    def test_content_tree_with_child_instruction(self):
        template_xml = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_example_invalid_template.odt'
        )
        t = Template(template_xml, get_secure_filename())
        usr_insts = t.get_user_instructions()
        assert usr_insts == ['for="item in items"', '/for',
                             'for="item in items', '2', '"', '/for']

    def __load_and_convert_template(self, path):
        template_xml = pkg_resources.resource_filename(
            'py3o.template',
            path
        )
        t = Template(template_xml, get_secure_filename())
        expressions = t.get_all_user_python_expression()
        py_expr = t.convert_py3o_to_python_ast(expressions)
        return py_expr

    def test_access_global_variable_inside_loop(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_access_global_variable_inside_loop.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'global_var': Mock(val=1),
            'my4list': [0, 1, 2, 3],
        }
        json_str = res.render(user_data)
        self.assertEquals(json_str, {
            'global_var': dict(val=1),
            'my4list': [0, 1, 2, 3],
        })

    def test_access_in_loop_variable(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_access_in_loop_variable.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'my2list': [0, 1, 2, 3, 4]
        }
        json_str = res.render(user_data)
        assert json_str == {'my2list': [0, 1, 2, 3, 4]}

    def test_access_global_variable(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_access_global_variable.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'var0': 0,
            'var1': 1,
            'var2': 2,
        }
        json_str = res.render(user_data)
        assert json_str == {'var0': 0, 'var1': 1, 'var2': 2}

    def test_iterable_with_global_attribute(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_iterable_with_global_attribute.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'foo': Mock(my2list=[0, 1, 2, 3, 4])
        }
        json_str = res.render(user_data)
        assert json_str == {'foo': {'my2list': [0, 1, 2, 3, 4]}}

    def test_two_for_list_on_same_attribute(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_two_for_list_on_same_attribute.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'foo': Mock(
                my2list=[0, 1, 2, 3, 4],
                my3list=[5, 6, 7, 8],
            )
        }
        json_str = res.render(user_data)
        assert json_str == {
            'foo': {
                'my2list': [0, 1, 2, 3, 4],
                'my3list': [5, 6, 7, 8],
            }
        }

    def test_access_in_loop_variable_with_attribute(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_access_in_loop_variable_with_attribute.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'my3list': [
                Mock(val=0),
                Mock(val=1),
                Mock(val=2),
            ]
        }
        json_str = res.render(user_data)
        assert json_str == {
            'my3list': [
                {'val': 0},
                {'val': 1},
                {'val': 2},
            ]
        }

    def test_access_in_loop_variable_with_multiple_attribute(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/'
            'py3o_access_in_loop_variable_with_multiple_attribute.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'my3list': [
                Mock(foo=Mock(val=0)),
                Mock(foo=Mock(val=1)),
                Mock(foo=Mock(val=2)),
            ]
        }
        json_str = res.render(user_data)
        assert json_str == {
            'my3list': [
                {'foo': {'val': 0}},
                {'foo': {'val': 1}},
                {'foo': {'val': 2}},
            ]
        }

    def test_access_parent_variable_in_nested_loop(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_access_parent_variable_in_nested_loop.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'my9list': [Mock(val=10), Mock(val=11)]
        }
        json_str = res.render(user_data)
        assert json_str == {
            'my9list': [
                {'val': 10},
                {'val': 11},
            ]
        }

    def test_access_parent_variable_in_nested_loop_with_attribute(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/'
            'py3o_access_parent_variable_in_nested_loop_with_attribute.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'my10list': [
                Mock(my_list=[
                    Mock(val=10),
                    Mock(val=11),
                ]),
                Mock(my_list=[
                    Mock(val=20),
                    Mock(val=21),
                    Mock(val=22),
                ]),
            ]
        }
        json_str = res.render(user_data)
        assert json_str == {
            'my10list': [
                {'my_list': [
                    {'val': 10},
                    {'val': 11},
                ]},
                {'my_list': [
                    {'val': 20},
                    {'val': 21},
                    {'val': 22},
                ]},
            ]
        }

    def test_access_variable_in_nested_loop(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/'
            'py3o_access_variable_in_nested_loop.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'my8list': [
                [10, 11, 12],
                [20, 21, 22, 23],
            ]
        }
        json_str = res.render(user_data)
        assert json_str == {
            'my8list': [
                [10, 11, 12],
                [20, 21, 22, 23]
            ]
        }

    def test_bad_user_data(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_access_global_variable_inside_loop.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {}
        with self.assertRaises(Py3oDataError):
            res.render(user_data)

    def test_test_statement(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_test_statement.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'var0': 42,
            'myvar': Mock(
                var0=43,
                var1='val',
            ),
        }
        json_dict = res.render(user_data)
        assert json_dict == {
            'var0': 42,
            'myvar': {
                'var0': 43,
                'var1': 'val',
            },
        }

    def test_double_loop_on_same_object(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_double_loop_on_same_object.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'varlist': [
                Mock(var0=42, var1=-42),
                Mock(var0=170, var1=-170),
            ]
        }
        json_dict = res.render(user_data)
        self.assertEqual(json_dict, {
            'varlist': [
                {'var0': 42, 'var1': -42},
                {'var0': 170, 'var1': -170},
            ]
        })

    def test_false_value(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/test_false_value.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        false_mock = Mock()
        false_mock.__nonzero__ = lambda x: False
        false_mock.__bool__ = lambda x: False

        user_data = {
            'false_value': false_mock,
        }
        json_dict = res.render(user_data)
        assert json_dict == {
            'false_value': u'',
        }

    def test_enumerate(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_enumerate.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {
            'mylist': [10, 9, 8, 7, 6],
        }
        json_dict = res.render(user_data)
        assert json_dict == {
            'mylist': [10, 9, 8, 7, 6],
        }

    def test_enumerate_attribute_access(self):
        """For loop on enumerate: only attributes used in body are extracted.
        Since mylist is given as an argument to enumerate, a known builtin
        function, Py3oConvertor should be able to recognize that val refers
        to the elements of mylist.
        """
        expressions = [
            'for="i, val in enumerate(mylist)"',
            'i',
            'val.var0',
            'val.var2',
            '/for'
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {'mylist': [
            Mock(var0='0', var1=1, var2=2.0),
            Mock(var0=0, var1=1.0, var2='2'),
            Mock(var0=0.0, var1='1', var2=2),
        ]}
        json_dict = res.render(user_data)
        self.assertEqual(json_dict, {'mylist': [
            dict(var0='0', var2=2.0),
            dict(var0=0, var2='2'),
            dict(var0=0.0, var2=2),
        ]})

    def test_unknown_iterable_call(self):
        """For loop on unknown function call: extract the entire object"""

        expressions = [
            'for="i, val in not_enumerate(mylist)"',
            'i',
            'val.var0',
            'val.var2',
            '/for'
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {'mylist': [
            Mock(var0='0', var1=1, var2=2.0),
            Mock(var0=0, var1=1.0, var2='2'),
            Mock(var0=0.0, var1='1', var2=2),
        ]}
        json_dict = res.render(user_data)
        self.assertEqual(json_dict, {'mylist': [
            user_data['mylist'][0],
            user_data['mylist'][1],
            user_data['mylist'][2],
        ]})

    def test_template_function_call(self):
        py_expr = self.__load_and_convert_template(
            'tests/templates/py3o_template_function_call.odt'
        )
        p = Py3oConvertor()
        res = p(py_expr)
        user_data = {
            'amount': 32.123,
        }
        json_dict = res.render(user_data)
        assert json_dict == {
            'amount': 32.123,
        }

    def test_template_function_call_in_for_loop(self):
        expressions = [
            'for="item in object.list"',
            'item.test',
            'format_number(item.var)',
            '/for',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        user_data = {'object': Mock(list=[
            Mock(var=32.123, test='2.3'),
            Mock(var=43.2, test='1.0')
        ])}
        json_dict = res.render(user_data)
        self.assertEqual(json_dict, {'object': {
            'list': [
                {'var': 32.123, 'test': '2.3'},
                {'var': 43.2, 'test': '1.0'},
            ]
        }})

    def test_double_for_loop_on_same_attribute(self):
        expressions = [
            'for="item in root.object.list"',
            'item.var1',
            '/for',
            'for="item in root.object.list"',
            'item.var0',
            '/for',
            '/for'
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)

        user_data = {'root': Mock(
            object=Mock(list=[
                Mock(var0=42, var1=-42),
                Mock(var0=170, var1=-170),
            ]))
        }
        json_dict = res.render(user_data)
        self.assertEqual(json_dict, {'root': {
            'object': {
                'list': [
                    dict(var0=42, var1=-42),
                    dict(var0=170, var1=-170),
                ]
            }
        }})

    def test_if(self):
        expressions = [
            'if="item.mytest"',
            'item.myvar',
            'item.myvar2',
            '/if',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'item': Mock(mytest=0, myvar=1, myvar2=2),
        })

        assert json_dict == {
            'item': {'mytest': 0, 'myvar': 1, 'myvar2': 2},
        }

    def test_if_global(self):
        expressions = [
            'if="mytest"',
            'item.myvar',
            'item.myvar2',
            '/if',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'item': Mock(myvar=1, myvar2=2),
            'mytest': 0
        })

        assert json_dict == {
            'mytest': 0,
            'item': {'myvar': 1, 'myvar2': 2},
        }

    def test_if_global_array(self):
        expressions = [
            'for="var in myarray"',
            'if="var"',
            'var',
            '/if',
            '/for',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'myarray': list(range(0, 5)),
        })

        assert json_dict == {
            'myarray': list(range(0, 5)),
        }

    def test_if_in_for(self):
        expressions = [
            'for="item in object.mylist"',
            'if="item.mytest"',
            'item.myvar',
            'item.myvar2',
            '/if',
            '/for',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'object': Mock(mylist=[
                Mock(mytest=0, myvar=1, myvar2=2),
                Mock(mytest=3, myvar=4, myvar2=5),
                Mock(mytest=6, myvar=7, myvar2=8),
            ])
        })

        assert json_dict == {
            'object': {
                'mylist': [
                    {'mytest': 0, 'myvar': 1, 'myvar2': 2},
                    {'mytest': 3, 'myvar': 4, 'myvar2': 5},
                    {'mytest': 6, 'myvar': 7, 'myvar2': 8},
                ]
            }
        }

    def test_if_comparator(self):
        expressions = [
            'if="a in b"',
            '/if'
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'a': 1,
            'b': [0, 1, 2],
        })

        assert json_dict == {
            'a': 1,
            'b': [0, 1, 2],
        }

    def test_if_comparator_literal(self):
        expressions = [
            'if="a == 1"',
            '/if'
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'a': 1,
        })

        self.assertEqual(json_dict, {
            'a': 1,
        })

    # Test function

    def test_function_in_for(self):
        expressions = [
            'for="item in object.mylist"',
            'format_date(item.date)',
            '/for',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'object': Mock(mylist=[
                Mock(date='2015-12-13'),
                Mock(date='2015-12-14'),
                Mock(date='2015-12-15'),
            ])
        })

        assert json_dict == {
            'object': {
                'mylist': [
                    {'date': '2015-12-13'},
                    {'date': '2015-12-14'},
                    {'date': '2015-12-15'},
                ]
            }
        }

    def test_function_keywords(self):
        expressions = [
            'myfunc(object.var, k=object.k)',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'object': Mock(var=0, k=1),
        })

        assert json_dict == {
            'object': {'var': 0, 'k': 1}
        }

    def test_function_array(self):
        expressions = [
            'for="var in myarray"',
            'myfunc(var)',
            '/for'
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        json_dict = res.render({
            'myarray': [0, 1, 2, 3],
        })

        assert json_dict == {
            'myarray': [0, 1, 2, 3],
        }

    def test_empty_for_loop(self):
        u"""Test ast extraction on for loops whose body do not use any data"""
        expressions = [
            'for="var in myarray"',
            '/for',
        ]
        py_expr = Template.convert_py3o_to_python_ast(expressions)
        self.assertEqual(py_expr.strip(), 'for var in myarray:\n pass')
        p = Py3oConvertor()
        res = p(py_expr)
        data = {'myarray': [0, 1, 2, 3]}
        json_dict = res.render(data)
        self.assertEqual(json_dict, data)

    def test_unpack_from_data_source(self):

        expressions = [
            'for="a, b in myarray"',
            'a',
            'b.c',
            '/for'
        ]

        py_expr = Template.convert_py3o_to_python_ast(expressions)
        p = Py3oConvertor()
        res = p(py_expr)
        data = {'myarray': [
            (1, Mock(c=2, d=3)),
            (4, Mock(c=5, d=6)),
        ]}
        json_dict = res.render(data)
        self.assertEqual(json_dict, {'myarray': [
            [1, {'c': 2}],
            [4, {'c': 5}],
        ]})



# -*- encoding: utf-8 -*-
import datetime
import os
import unittest
import zipfile
import traceback
import copy
import base64

import lxml.etree
import pkg_resources
import six

from io import BytesIO

from genshi.template import TemplateError
from pyjon.utils import get_secure_filename

from py3o.template import Template, TextTemplate, TemplateException
from py3o.template.main import XML_NS, get_soft_breaks

if six.PY3:
    # noinspection PyUnresolvedReferences
    from unittest.mock import Mock
elif six.PY2:
    # noinspection PyUnresolvedReferences
    from mock import Mock


class TestTemplate(unittest.TestCase):

    def tearDown(self):
        pass

    def setUp(self):
        pass

    def test_example_1(self):
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_example_template.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)
        template.set_image_path(
            'staticimage.logo',
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/images/new_logo.png'
            )
        )

        class Item(object):
            pass

        items = list()

        item1 = Item()
        item1.val1 = 'Item1 Value1'
        item1.val2 = 'Item1 Value2'
        item1.val3 = 'Item1 Value3'
        item1.Currency = 'EUR'
        item1.Amount = '12345.35'
        item1.InvoiceRef = '#1234'
        items.append(item1)

        # if you are using python 2.x you should use xrange
        for i in range(1000):
            item = Item()
            item.val1 = 'Item%s Value1' % i
            item.val2 = 'Item%s Value2' % i
            item.val3 = 'Item%s Value3' % i
            item.Currency = 'EUR'
            item.Amount = '6666.77'
            item.InvoiceRef = 'Reference #%04d' % i
            items.append(item)

        document = Item()
        document.total = '9999999999999.999'

        data = dict(items=items, document=document)
        error = False
        try:
            template.render(data)
        except ValueError as e:
            print('The template did not render properly...')
            traceback.print_exc()
            error = True

        assert error is False

    def test_list_duplicate(self):
        """test duplicated listed get a unique id"""
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_list_template.odt'
        )
        outname = get_secure_filename()

        template = Template(template_name, outname)

        class Item(object):
            def __init__(self, val):
                self.val = val
        data_dict = {
            "items": [Item(1), Item(2), Item(3), Item(4)]
        }

        error = False

        template.set_image_path(
            'staticimage.logo',
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/images/new_logo.png'
            )
        )
        template.render(data_dict)

        outodt = zipfile.ZipFile(outname, 'r')
        try:
            content_list = [
                lxml.etree.parse(BytesIO(outodt.read(filename)))
                for filename in template.templated_files
            ]
        except lxml.etree.XMLSyntaxError as e:
            error = True
            print(
                "List was not deduplicated->{}".format(e)
            )

        # remove end file
        os.unlink(outname)

        assert error is False

        # first content is the content.xml
        content = content_list[0]
        list_expr = '//text:list'
        list_items = content.xpath(
            list_expr,
            namespaces=template.namespaces
        )
        ids = []
        for list_item in list_items:
            ids.append(
                list_item.get(
                    '{}id'.format(XML_NS)
                )
            )
        assert ids, "this list of ids should not be empty"
        assert len(ids) == len(set(ids)), "all ids should have been unique"

    def test_missing_opening(self):
        """test orphaned /for raises a TemplateException"""
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_missing_open_template.odt'
        )
        outname = get_secure_filename()
        try:
            template = Template(template_name, outname)

        finally:
            os.remove(outname)

        class Item(object):
            def __init__(self, val):
                self.val = val
        data_dict = {
            "items": [Item(1), Item(2), Item(3), Item(4)]
        }

        template.set_image_path(
            'staticimage.logo',
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/images/new_logo.png'
            )
        )
        # this will raise a TemplateException... or the test will fail
        error_occured = False
        try:
            template.render(data_dict)

        except TemplateException as e:
            error_occured = True
            # make sure this is the correct TemplateException that pops
            assert e.message == "No open instruction for /for"

        # and make sure we raised
        assert error_occured is True

    def test_ignore_undefined_variables_logo(self):

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_logo.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        data = {}

        error = True
        try:
            template.render(data)
            print("Error: template contains a logo variable that must be "
                  "replaced")
        except ValueError:
            error = False

        assert error is False

        template = Template(template_name, outname,
                            ignore_undefined_variables=True)

        error = False
        try:
            template.render(data)
        except:
            traceback.print_exc()
            error = True

        assert error is False

    def test_ignore_undefined_variables_1(self):

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_undefined_variables_1.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        data = {}

        error = True
        try:
            template.render(data)
            print("Error: template contains variables that must be "
                  "replaced")
        except TemplateError:
            error = False

        assert error is False

        template = Template(template_name, outname,
                            ignore_undefined_variables=True)

        error = False
        try:
            template.render(data)
        except:
            traceback.print_exc()
            error = True

        assert error is False

    def test_ignore_undefined_variables_2(self):
        """
        Test ignore undefined variables for template with dotted variables like
        py3o.document.value
        """

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_undefined_variables_2.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        data = {}

        error = True
        try:
            template.render(data)
            print("Error: template contains variables that must be "
                  "replaced")
        except TemplateError:
            error = False

        assert error is False

        template = Template(template_name, outname,
                            ignore_undefined_variables=True)

        error = True
        try:
            template.render(data)
            print("Error: template contains dotted variables that must be "
                  "replaced")
        except TemplateError:
            error = False

        assert error is False

    def test_invalid_template_1(self):
        """a template should not try to define a /for and a for on the same
        paragraph
        """

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_example_invalid_template.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        class Item(object):
            pass

        items = list()

        item1 = Item()
        item1.val1 = 'Item1 Value1'
        item1.val2 = 'Item1 Value2'
        item1.val3 = 'Item1 Value3'
        item1.Currency = 'EUR'
        item1.Amount = '12345.35'
        item1.InvoiceRef = '#1234'
        items.append(item1)

        # if you are using python 2.x you should use xrange
        for i in range(1000):
            item = Item()
            item.val1 = 'Item%s Value1' % i
            item.val2 = 'Item%s Value2' % i
            item.val3 = 'Item%s Value3' % i
            item.Currency = 'EUR'
            item.Amount = '6666.77'
            item.InvoiceRef = 'Reference #%04d' % i
            items.append(item)

        document = Item()
        document.total = '9999999999999.999'

        data = dict(
            items=items,
            items2=copy.copy(items),
            document=document
        )

        error = False
        try:
            template.render(data)
        except TemplateException:
            error = True

        assert error is True, "This template should have been refused"

    def test_template_with_function_call(self):
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_template_function_call.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        data_dict = {
            'amount': 32.123,
        }

        template.render(data_dict)
        outodt = zipfile.ZipFile(outname, 'r')

        content_list = lxml.etree.parse(
            BytesIO(outodt.read(template.templated_files[0]))
        )

        result_a = lxml.etree.tostring(
            content_list,
            pretty_print=True,
        ).decode('utf-8')

        result_e = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/template_with_function_call_result.xml'
            )
        ).read()

        result_a = result_a.replace("\n", "").replace(" ", "")
        result_e = result_e.replace("\n", "").replace(" ", "")

        assert result_a == result_e

    def test_format_date(self):
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_template_format_date.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        data_dict = {
            'datestring': '2015-08-02',
            'datetimestring': '2015-08-02 17:05:06',
            'datestring2': '2015-10-15',
            'datetime': datetime.datetime.strptime(
                '2015-11-13 17:00:20',
                '%Y-%m-%d %H:%M:%S'
            ),
        }

        template.render(data_dict)
        outodt = zipfile.ZipFile(outname, 'r')

        content_list = lxml.etree.parse(
            BytesIO(outodt.read(template.templated_files[0]))
        )

        result_a = lxml.etree.tostring(
            content_list,
            pretty_print=True,
        ).decode('utf-8')

        result_e = open(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/template_format_date_result.xml'
            )
        ).read()

        result_a = result_a.replace("\n", "").replace(" ", "")
        result_e = result_e.replace("\n", "").replace(" ", "")

        assert result_a == result_e

    def test_format_date_exception(self):
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_template_format_date_exception.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        data_dict = {
            'date': '2015/08/02',
        }

        # this will raise a TemplateException... or the test will fail
        error_occured = False
        try:
            template.render(data_dict)

        except TemplateException as e:
            error_occured = True

        # and make sure we raised
        assert error_occured is True

    def test_style_application_with_function_call(self):
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/style_application_with_function_call.odt'
        )

        outname = get_secure_filename()

        template = Template(template_name, outname)

        data_dict = {
            'date': '2015-08-02',
        }

        template.render(data_dict)
        outodt = zipfile.ZipFile(outname, 'r')

        content_list = lxml.etree.parse(
            BytesIO(outodt.read(template.templated_files[0]))
        )

        result_a = lxml.etree.tostring(
            content_list,
            pretty_print=True,
        ).decode('utf-8')

        result_e = open(
            pkg_resources.resource_filename(
                'py3o.template', (
                    'tests/templates/'
                    'style_application_with_function_call_result.xml'
                )
            )
        ).read()

        result_a = result_a.replace("\n", "").replace(" ", "")
        result_e = result_e.replace("\n", "").replace(" ", "")

        assert result_a == result_e

    def test_image_injection(self):
        """Test insertion of images from the data source into the template"""

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_image_injection.odt'
        )
        logo_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/images/new_logo.png'
        )
        image_names = [
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/images/image{i}.png'.format(i=i)
            ) for i in range(1, 4)
        ]
        outname = get_secure_filename()

        template = Template(template_name, outname)
        logo = open(logo_name, 'rb').read()
        images = [open(iname, 'rb').read() for iname in image_names]

        data_dict = {
            'items': [
                Mock(val1=i, val3=i ** 2, image=base64.b64encode(image))
                for i, image in enumerate(images, start=1)
            ],
            'document': Mock(total=6),
            'logo': logo,
        }

        template.render(data_dict)
        outodt = zipfile.ZipFile(outname, 'r')

        content_list = lxml.etree.parse(
            BytesIO(outodt.read(template.templated_files[0]))
        )
        namelist = outodt.namelist()

        i = 0
        nmspc = template.namespaces
        table = content_list.find('//table:table', nmspc)
        frame_path = 'table:table-cell/text:p/draw:frame'
        for row in table.findall('table:table-row', nmspc):

            frame_elem = row.find(frame_path, nmspc)
            if frame_elem is None:
                continue
            image_elem = frame_elem.find('draw:image', nmspc)
            self.assertIsNotNone(image_elem)

            href = image_elem.get('{{{}}}href'.format(nmspc['xlink']))
            self.assertTrue(href)
            self.assertIn(href, namelist)
            self.assertEqual(images[i], outodt.read(href))

            frame_elem.remove(image_elem)
            i += 1

        self.assertEqual(i, 3, u"Images were not found in the output")

        expected_xml = lxml.etree.parse(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/image_injection_result.xml'
            )
        )
        result = lxml.etree.tostring(
            content_list, pretty_print=True,
        ).decode('utf-8')
        expected = lxml.etree.tostring(
            expected_xml, pretty_print=True,
        ).decode('utf-8')
        result = result.replace("\n", "").replace(" ", "")
        expected = expected.replace("\n", "").replace(" ", "")

        self.assertEqual(result, expected)

    def test_ignore_undefined_variables_image_injection(self):
        """Test ignore undefined variables for injected image"""

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_image_injection.odt'
        )

        outname = get_secure_filename()

        data = {
            'items': [],
            'document': Mock(total=6),
        }

        template = Template(template_name, outname)
        error = True
        try:
            template.render(data)
            print("Error: template contains variables that must be "
                  "replaced")
        except TemplateError:
            error = False

        self.assertFalse(error)

        template = Template(
            template_name, outname, ignore_undefined_variables=True
        )
        error = False
        try:
            template.render(data)
        except:
            traceback.print_exc()
            error = True

        self.assertFalse(error)

    def test_text_template(self):

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_text_template'
        )

        user_data = {'mylist': [
            Mock(var0=1, var1='1', var2=1.0),
            Mock(var0=2, var1='2', var2=2.0),
            Mock(var0=3, var1='3', var2=3.0),
        ]}

        outname = get_secure_filename()

        template = TextTemplate(template_name, outname)
        template.render(user_data)
        result = open(outname, 'rb').read()

        expected = u''.join(
            u'{} {} {}\n'.format(line.var0, line.var1, line.var2)
            for line in user_data['mylist']
        ).encode('utf-8')

        self.assertEqual(result, expected)

    def test_ignore_undefined_variables_text_template(self):

        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_text_template'
        )

        user_data = {}

        outname = get_secure_filename()

        template = TextTemplate(template_name, outname)
        error = True
        try:
            template.render(user_data)
            print("Error: template contains variables that must be "
                  "replaced")
        except TemplateError:
            error = False

        self.assertFalse(error)

        template = TextTemplate(
            template_name, outname, ignore_undefined_variables=True
        )
        error = False
        try:
            template.render(user_data)
        except:
            traceback.print_exc()
            error = True

        self.assertFalse(error)

    def test_remove_soft_page_breaks(self):
        template_xml = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_soft_page_break.odt'
        )
        outname = get_secure_filename()

        template = Template(template_xml, outname)
        soft_breaks = get_soft_breaks(
            template.content_trees[0], template.namespaces
        )
        self.assertEqual(len(soft_breaks), 2)

        template.remove_soft_breaks()
        soft_breaks = get_soft_breaks(
            template.content_trees[0], template.namespaces
        )
        self.assertEqual(len(soft_breaks), 0)

        template = Template(template_xml, outname)
        soft_breaks = get_soft_breaks(
            template.content_trees[0], template.namespaces
        )
        self.assertEqual(len(soft_breaks), 2)

        template.render(data={"list1": [1, 2, 3]})
        soft_breaks = get_soft_breaks(
            template.content_trees[0], template.namespaces
        )
        self.assertEqual(len(soft_breaks), 0)

        outodt = zipfile.ZipFile(outname, 'r')
        content_list = lxml.etree.parse(
            BytesIO(outodt.read(template.templated_files[0]))
        )

        nmspc = template.namespaces
        paragraphs = content_list.findall('//text:p', nmspc)
        bottom_break_paragraphs, middle_break_paragraphs = 0, 0
        for p in paragraphs:
            if not p.text:
                continue
            text = p.text.strip()
            if text == (
                u"This is a text with a margin at the bottom and a "
                u"soft-page-break"
            ):
                bottom_break_paragraphs += 1
            elif text == (
                u"This is a paragraph that is cut in half by a "
                u"soft-page-break. This text should not remain cut "
                u"in half after rendering."
            ):
                middle_break_paragraphs += 1
            else:
                self.fail(u"Unidentified text in result: {}".format(text))

        self.assertEqual(bottom_break_paragraphs, 3)
        self.assertEqual(middle_break_paragraphs, 3)

    def test_remove_soft_breaks_without_tail(self):
        template_xml = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_page_break_without_tail.odt'
        )
        t = Template(template_xml, get_secure_filename())
        soft_breaks = get_soft_breaks(t.content_trees[0], t.namespaces)
        assert len(soft_breaks) > 0

        t.remove_soft_breaks()
        soft_breaks = get_soft_breaks(t.content_trees[0], t.namespaces)
        assert len(soft_breaks) == 0

        t = Template(template_xml, get_secure_filename())
        soft_breaks = get_soft_breaks(t.content_trees[0], t.namespaces)
        assert len(soft_breaks) > 0

        t.render(data={"items": [
            {'Amount': 3, 'Currency': 'D'},
            {'Amount': 2, 'Currency': 'E'},
            {'Amount': 1, 'Currency': 'C'},
        ]})
        soft_breaks = get_soft_breaks(t.content_trees[0], t.namespaces)
        assert len(soft_breaks) == 0

    def test_invalid_links(self):
        u"""Check that exceptions are raised on link url and text mismatch"""

        templates = [
            ('py3o_invalid_link.odt', 'url and text do not match.*'),
            ('py3o_invalid_link_old.odt', 'url and text do not match.*'),
            ('py3o_invalid_link_none.odt', 'Link text not found'),
        ]

        for template, error in templates:
            template_fname = pkg_resources.resource_filename(
                'py3o.template', 'tests/templates/{}'.format(template)
            )
            t = Template(template_fname, get_secure_filename())
            with self.assertRaisesRegexp(TemplateException, error):
                t.render({'amount': 0.0})

    def test_table_cell_function_call(self):
        u"""Test function calls inside ODT table cells"""
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_table_cell_function_call.odt'
        )
        outname = get_secure_filename()
        template = Template(template_name, outname)

        data_dict = {
            'items': [
                Mock(val1=i, val2=range(i), val3=i ** 2)
                for i in range(1, 4)
                ],
            'document': Mock(total=6),
        }

        template.render(data_dict)

    def test_table_cell_for_loop(self):
        u"""Test for loop inside ODT table cells"""
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_table_cell_for_loop.odt'
        )
        outname = get_secure_filename()
        template = Template(template_name, outname)

        data_dict = {
            'items': [
                Mock(val1=i, val2=range(i), val3=i ** 2)
                for i in range(1, 4)
            ],
            'document': Mock(total=6),
        }

        template.render(data_dict)

    def test_odt_value_styles(self):
        u"""Test odf_value attribute and ODT styles"""
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_odt_value_styles.odt'
        )
        outname = get_secure_filename()
        template = Template(template_name, outname)

        data_dict = {
            'string_date': '1999-12-30',
            'odt_value_date': Mock(
                __str__=lambda s: '2009-07-06',
                odf_value=40000,
                odf_type='date',
            )
        }

        template.render(data_dict)
        outodt = zipfile.ZipFile(outname, 'r')
        content_list = lxml.etree.parse(
            BytesIO(outodt.read(template.templated_files[0]))
        )

        expected_xml = lxml.etree.parse(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/odt_value_styles_result.xml'
            )
        )
        result = lxml.etree.tostring(
            content_list, pretty_print=True,
        ).decode('utf-8')
        expected = lxml.etree.tostring(
            expected_xml, pretty_print=True,
        ).decode('utf-8')
        result = result.replace("\n", "").replace(" ", "")
        expected = expected.replace("\n", "").replace(" ", "")

        self.assertEqual(result, expected)

    def test_ods_value_styles(self):
        u"""Test odf_value attribute and ODS styles"""
        template_name = pkg_resources.resource_filename(
            'py3o.template',
            'tests/templates/py3o_ods_value_styles.ods'
        )
        outname = get_secure_filename()
        template = Template(template_name, outname)

        data_dict = {
            'string_date': '1999-12-30',
            'odf_value_date': Mock(
                __str__=lambda s: '2009-07-06',
                odf_value=40000,
                odf_type='date',
            )
        }

        template.render(data_dict)
        outodt = zipfile.ZipFile(outname, 'r')
        content_list = lxml.etree.parse(
            BytesIO(outodt.read(template.templated_files[0]))
        )

        expected_xml = lxml.etree.parse(
            pkg_resources.resource_filename(
                'py3o.template',
                'tests/templates/ods_value_styles_result.xml'
            )
        )
        result = lxml.etree.tostring(
            content_list, pretty_print=True,
        ).decode('utf-8')
        expected = lxml.etree.tostring(
            expected_xml, pretty_print=True,
        ).decode('utf-8')
        result = result.replace("\n", "").replace(" ", "")
        expected = expected.replace("\n", "").replace(" ", "")

        self.assertEqual(result, expected)



# -*- encoding: utf-8 -*-
"""py3o.template exposes a dirt simple API to render templated OpenOffice
documents into real OpenOffice documents with all your data merged-in.
"""

from py3o.template.main import Template
from py3o.template.main import TextTemplate
from py3o.template.main import TemplateException



from py3o.template import Template

t = Template(
    "py3o_example_template_single_cell.odt",
    "py3o_example_output_single_cell.odt"
)

t.set_image_path('staticimage.logo', 'images/new_logo.png')


class Item(object):
    pass

items = list()

item1 = Item()
item1.val1 = 'Item1 Value1'
item1.val2 = 'Item1 Value2'
item1.val3 = 'Item1 Value3'
item1.Currency = 'EUR'
item1.Amount = '12,345.35'
item1.InvoiceRef = '#1234'
items.append(item1)

for i in xrange(1000):
    item = Item()
    item.val1 = 'Item%s Value1' % i
    item.val2 = 'Item%s Value2' % i
    item.val3 = 'Item%s Value3' % i
    item.Currency = 'EUR'
    item.Amount = '6,666.77'
    item.InvoiceRef = 'Reference #%04d' % i
    items.append(item)

document = Item()
document.total = '9,999,999,999,999.999'

data = dict(items=items, document=document)
t.render(data)



"""This file contains all the data structures used by Py3oConvertor
See the docstring of Py3oConvertor.__call__() for further information
"""
from numbers import Number


class Py3oDataError(Exception):
    pass


class Py3oObject(dict):
    """ Base class to be inherited.
    """

    is_list = False

    def render(self, data):  # pragma: no cover
        raise NotImplementedError("This function should be overriden")

    def __repr__(self):  # pragma: no cover
        res = super(Py3oObject, self).__repr__()
        return "{}({})".format(
            self.__class__.__name__,
            res
        )

    def get_size(self):
        """Return the max depth of the object
        """
        sizes = [val.get_size() for val in self.values()]
        if not sizes:
            return 0
        return max(sizes) + 1

    def get_key(self):
        """Return the first key
        """
        return next(iter(self.keys()))

    def get_tuple(self):
        """Return the value of the Py3oObject as a tuple.
        As a default behavior, the object returns None.
        """
        return None

    def unpack(self, target):
        self_tup = self.get_tuple()
        target_tup = target.get_tuple()
        if target_tup is None:
            yield (target, self)
        elif self_tup is None:
            yield (None, self)
            array = self[self.get_key()]
            array.is_list = True
            for i, target_elem in enumerate(target_tup):
                self[self.get_key()] = tmpname = Py3oName({i: Py3oName()})
                yield (target_elem, self)
                array[i] = tmpname[i]
            self[self.get_key()] = array
        else:
            diff = len(target_tup) - len(self_tup)
            if diff != 0:  # pragma: no cover
                raise ValueError(u"Unpack Error: {} != {}".format(
                    target_tup, self_tup
                ))
            for t in zip(target_tup, self_tup):
                yield t

    def rupdate(self, other):
        """Update recursively the Py3oObject self with the Py3oObject other.
        Example:
        self = Py3oObject({
            'a': Py3oObject({}),
            'b': Py3oObject({
                'c': Py3oObject({}),
            }),
        })
        other = Py3oObject({
            'b': Py3oObject({
                'd': Py3oObject({}),
            }),
        })
        res = Py3oObject({
            'a': Py3oObject({}),
            'b': Py3oObject({
                'c': Py3oObject({}),
                'd': Py3oObject({}),
            }),
        })
        """
        for key, value in other.items():
            if key in self:
                self[key].rupdate(value)
            else:
                self[key] = value

    def rget(self, other):
        """Get the value for the path described by the other Py3oObject.

        Recursively checks that the values in other can be found in self.

        The method returns the values of self and other at the point where
        the search stopped.
        If other is a leaf, the search stops sucessfully. The method returns
        True, the value that corresponds to the path described by other, and
        the leaf in question.
        If other cannot be found in self, the search stops unsuccessfully.
        The method returns False, the value that corresponds to the deepest
        point reached in self, and the rest of the path.

        Example:
        self = Py3oObject({
            'a': Py3oObject({}),
            'b': Py3oObject({
                'c': Py3oObject({}),
            }),
        })
        other = Py3oObject({
            'b': Py3oObject({
                'd': Py3oObject({}),
            }),
        })
        res = (
            False,
            Py3oObject({'c': Py3oObject({})}),  # is self['b']
            Py3oObject({'d': Py3oObject({})}),  # is other['b']
        )
        if other['b'] was a leaf, res[0] would be True and res[2] the leaf.

        :return: A triple:
          - True if the search was successful, False otherwise
          - The active sub-element of self when the search stopped
          - The active sub-element of other when the search stopped
        """
        if not other:
            return True, self, other
        other_key = other.get_key()
        if other_key in self:
            return self[other_key].rget(other[other_key])
        else:
            return False, self, other

    def render_children(self, data):
        if self.is_list:
            res = [self[i].render(item) for i, item in enumerate(data)]
        else:
            res = {
                key: value.render(getattr(data, key))
                for key, value in self.items()
            }
        return res


class Py3oModule(Py3oObject):
    def render(self, data):
        """ This function will render the datastruct according
         to the user's data
        """
        res = {}
        for key, value in self.items():
            subdata = data.get(key, None)
            if subdata is None:
                raise Py3oDataError(
                    "The key '%s' must be present"
                    " in your data dictionary" % key
                )
            # Spread only the appropriate data to its children
            val = value.render(data.get(key))
            if val is not None:
                res[key] = val
        return res


class Py3oArray(Py3oObject):
    """ A class representing an iterable value in the data structure.
    The attribute direct_access will tell if this class should be considered
     as a list of dict or a list of values.
    """
    def __init__(self):
        super(Py3oArray, self).__init__()
        self.direct_access = False

    def render(self, data):
        """ This function will render the datastruct according
        to the user's data
        """
        if self.direct_access:
            res = data
        elif not self:  # pragma: no cover
            res = None
        else:
            res = [self.render_children(d) for d in data]
        return res


class Py3oName(Py3oObject):
    """ This class holds information of variables.
    Keys are attributes and values the type of this attribute
     (another Py3o class or a simple value)
    i.e.: i.egg -> Py3oName({'i': Py3oName({'egg': Py3oName({})})})
    """
    def render(self, data):
        """ This function will render the datastruct according
        to the user's data
        """
        if not self:
            # We only send False values if the value is a number
            # otherwise we convert the False into an empty string
            res = data if data or isinstance(data, Number) else u""
        else:
            res = self.render_children(data)
        return res


class Py3oCall(Py3oObject):
    """This class holds information of function call.
    'name' holds the name of function as a Py3oName
    The keys are the arguments as:
        - numeric keys are positional arguments ordered ascendently
        - string keys are keywords arguments
    """

    return_format = None

    def __init__(self, name, dict):
        super(Py3oCall, self).__init__(dict)
        self.name = name

    def get_tuple(self):  # pragma: no cover
        raise SyntaxError(u"Can't assign to function call")

    def unpack(self, target):
        target_tup = target.get_tuple()
        args = set(self.keys())

        if self.return_format is None:
            # Generic/unknown function, do not make assumptions about targets
            res = [(target, None) for target in target_tup]
            res += [(None, self[arg]) for arg in args]

        elif isinstance(self.return_format, tuple):
            # Bind targets to function arguments according to return_format
            return_value = []
            for return_var in self.return_format:
                if return_var is None:
                    return_value.append(Py3oDummy())
                else:
                    return_value.append(self.get(return_var, None))

            if len(return_value) == len(target_tup):
                res = zip(target_tup, return_value)

            # TODO: Manage Py3oContainer values inside for loop body
            # elif len(target_tup) == 1:
            #     res = [(target_tup[0], Py3oContainer(return_value))]
            #     res += [(self[arg], None) for arg in args]

            else:  # pragma: no cover
                raise ValueError(u"Unpack Error")

        else:
            # Single return value is bound to one of the arguments.
            res = [(target, self.get(self.return_format, None))]

        for tup in res:
            yield tup


class Py3oEnumerate(Py3oCall):
    """Represent an enumerate call"""
    return_format = (None, 0)


class Py3oContainer(Py3oObject):
    """Represent a container defined in the template.
    This container can be:
    _ A literal list, tuple, set or dict definition
    _ A tuple of variables that are the target of an unpack assignment
    """
    def __init__(self, values):
        super(Py3oContainer, self).__init__()
        self.values = values

    def get_tuple(self):
        """Return the container's values in a tuple"""
        return tuple(self.values)


class Py3oDummy(Py3oObject):
    """ This class holds temporary dict, or unused attribute
     such as counters from enumerate()
    """
    pass


class Py3oBuiltin(Py3oObject):
    """ This class holds information about builtins
    """

    builtins = {
        'enumerate': Py3oEnumerate
    }

    @classmethod
    def from_name(cls, name=None):
        """Return the Py3oObject subclass for the given built-in name
        Return None if the name does not correspond to a builtin.

        :param name: A Py3oObject instance that represent a name/attribute path
        :return: A Py3oObject subclass or None
        """
        name_key = name.get_key()
        builtin = cls.builtins.get(name_key)
        # TODO: uncomment to implement standard modules in cls.builtins
        # if builtin and name.get_size() > 1:
        #     if isinstance(builtin, Py3oBuiltin):
        #         builtin = builtin.from_name(name[name_key])
        #     else:
        #     builtin = None
        return builtin



from py3o.template import Template

t = Template("asimage.odt", "asimage_output.odt")

# t.set_image_path('logo', 'images/new_logo.png')


class Item(object):
    pass

item = Item()
item.image = open('images/new_logo.png', 'rb').read()
item.image2 = open('images/new_logo.png', 'rb').read()
item.image3 = open('images/new_logo.png', 'rb').read()

data = {"object": item}
t.render(data)



# -*- encoding: utf-8 -*-
from py3o.template import Template

t = Template("simple_calc.ods", "calc.ods")


class Item(object):
    pass

items = list()

# if you are using python 2.x you should use xrange
for i in range(124):
    item = Item()
    item.col1 = 'Item%s Value1' % i
    item.col2 = i
    item.col3 = 'type%s' % (i % 2)
    item.col4 = 'Some description that never changes'

    item.image = open('images/dot%s.png' % (i % 2), 'rb').read()

    items.append(item)

data = dict(items=items)
t.render(data)



# -*- encoding: utf-8 -*-
from setuptools import setup, find_packages
import codecs

version = '0.9.10'

setup(
    name='py3o.template',
    version=version,
    description="An easy solution to design reports using OpenOffice",
    long_description=codecs.open(
        "README.rst", mode='r', encoding="utf-8"
    ).read(),
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords='LibreOffice OpenOffice templating PDF',
    author='Florent Aide',
    author_email='florent.aide@gmail.com',
    url='http://bitbucket.org/faide/py3o.template',
    license='MIT License',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=['py3o'],
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'setuptools',
        'six >= 1.9',
        'lxml',
        'genshi >= 0.7',
        'pyjon.utils > 0.6',
    ],
    entry_points="""
    # -*- Entry points: -*-
    """,
    tests_require=['nose', 'nosexcover', 'mock'],
    test_suite='nose.collector',
)
